--[[---------------------------------------
	VARIABLE CACHING
-----------------------------------------]]

-- Lua Functions
local pairs, print, select, tonumber, tostring, type, unpack = pairs, print, select, tonumber, tostring, type, unpack
local find, format, lower = string.find, string.format, string.lower
-- WoW API Functions / Variables
local GetCVar = GetCVar
local IsShiftKeyDown = IsShiftKeyDown
local PlaySoundFile = PlaySoundFile


local SCA, Ace3, LSM, GUI = unpack(select(2,...))

function Ace3:Spacer(order,width,db,expectedSource)
	local spacing = {
		order = order,
		type = "description",
		name = "",
		hidden = function()
			if (db) then
				if (db.selected ~= "empty" and db[db.selected].options.source == expectedSource) then
					return true
				else
					return false
				end
			else
				return false
			end
		end,
		width = width,
	}
	
	return spacing
end

--[[-------------------------------------------
	
	OPTION TOGGLE TOOLS
	
---------------------------------------------]]
function Ace3:TutorialToggle(db,order,width)
	local tutorial = {
		order = order,
		type = "toggle",
		name = "|cFF94FF5FTutorial|r",
		desc = "Displays tutorial panels to help you assign sounds to specific spells.\n\n|cFF00FF00Right-clicking a panel will hide it|r",
		get = function()
			return db.tutorial.enabled
		end,
		set = function(self,value)
			db.tutorial.enabled = value
			Ace3:UpdateTutorial(db)
			
			if (not value) then
				for k,_ in pairs(db.tutorial.flags) do
					db.tutorial.flags[k] = false
				end
			end
		end,
		width = width,
	}
	
	return tutorial
end

function Ace3:EventToggle(db,event,order,width,title,desc,tab,isSpecific,noSeparate,noTutorial)
	local event = {
		order = order,
		type = "toggle",
		name = title,
		desc = desc,
		disabled = function()
			if (isSpecific) then
				if (db.validSpell) then
					if (Ace3:CheckSpell(event,db.validSpell)) then
						return false
					else
						return true
					end
				else
					return true
				end
			else
				return false
			end
		end,
		get = function() 
			return db.events[event]
		end,
		set = function(self,value)
			db.events[event] = value
			
			if (title == "Damage" or title == "Healing") then
				if (SCA.temp.errors[lower(title)]) then
					SCA.temp.errors[lower(title)] = nil
				end
			end
			
			if (not noSeparate) then
				if (isSpecific) then
					if (value and Ace3:CheckSpell(event,db.validSpell)) then
						db.numEvents = db.numEvents + 1
					else
						db.numEvents = db.numEvents - 1
					end
				else
					if (value) then
						db.numEvents = db.numEvents + 1
					else
						db.numEvents = db.numEvents - 1
					end
				end
				
				if (db.isSeparate and db.numEvents <= 1) then
					db.isSeparate = false
				end
				--Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific)
			end
			
			if (not noSeparate or noSeparate == "partial") then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
			
			if (not noTutorial) then
				db.tutorial.flags.spellEvents = true
				Ace3:UpdateTutorial(db,'toggle')
			end
			
			if (db.selected == "empty") then
				Ace3:NavigateInterfaceOptions(SCA.temp.selectedMenuItem,true)
			end
		end,
		width = width,
	}
	
	return event
end

--[[-------------------------------------------
	
	REPORTING TOOLS
	
---------------------------------------------]]
function Ace3:DisabledError(db,event)
	local disabledError = {
		order = 0,
		type = "description",
		hidden = function()
			if (not event) then
				if (db.enabled) then
					return true
				else
					return false
				end
			else
				if (db.events[db.selected]) then
					return true
				else
					return false
				end
			end
		end,
		name = "|cFFFF8585CURRENTLY DISABLED IN QUICK ACCESS|r",
		width = "full",
	}
	
	return disabledError
end

function Ace3:SoundList(db,order,isSpecific,noSeparate)
	local soundList = {
		name = "Sound List",
		order = order,
		type = "group",
		dialogInline = true,
		hidden = function()
			if (isSpecific) then
				if (SCA.temp.addSound and db.isSeparate) then
					return false
				else
					return true
				end
			else
				if (not noSeparate) then
					if (db.isSeparate) then
						return false
					else
						return true
					end
				else
					if ((db.numEvents or 1) > 0) then
						return false
					else
						return true
					end
				end
			end
		end,
		args = {
			messages = {
				order = 1,
				type = "description",
				name = " ",
			},
		},
	}
	
	return soundList
end

function Ace3:Errors(dbName,order,isSpecific)
	local errors = {
		name = "Errors",
		order = order,
		type = "group",
		dialogInline = true,
		hidden = function()
			if (isSpecific) then
				if (SCA.temp.addSound and SCA.temp.errors[dbName]) then
					return false
				else
					return true
				end
			else
				if (SCA.temp.errors[dbName]) then
					return false
				else
					return true
				end
			end
		end,
		args = {
			messages = {
				order = 1,
				type = "description",
				name = " ",
			},
		},
	}
	
	return errors
end

--[[-------------------------------------------
	
	SOUND CONFIG TOOLS
	
---------------------------------------------]]
function Ace3:Source(db,errorName,values,isSpecific,noSeparate,noTutorial)
	
	local source = {
		order = 1,
		type = "select",
		name = "Audio Source",
		desc = "Choose which sound pack you want to play for crit sounds",
		--[[disabled = function()
			if (isSpecific) then
				if (not db.validSpell or db.numEvents == 0) then
					return true
				else
					return false
				end
			else
				if (not noSeparate) then
					if (db.numEvents == 0) then
						return true
					else
						return false
					end
				else
					return false
				end
			end
		end,]]
		get = function()
			local val
			
			if (db.selected == "empty") then
				val = "packs"
			else
				val = db[db.selected].options.source
			end
			
			return val
		end,
		set = function(self,value)
			if (SCA.temp.errors[errorName]) then
				SCA.temp.errors[errorName] = nil
			end
			
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (isSpecific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								db[k].options.source = value
							end
						else
							if (v) then
								db[k].options.source = value
							end
						end
					end
				else
					db[db.selected].options.source = value
				end
			else
				db[db.selected].options.source = value
			end
			
			if (not noTutorial) then
				db.tutorial.flags.source = true
				Ace3:UpdateTutorial(db)
			end
		end,
		values = values,
	}
	
	return source
end

function Ace3:Channel(db,values,isSpecific,noSeparate)
	channel = {
		order = 2,
		type = "select",
		name = "Sound Channel",
		desc = "Choose which channel for the audio to play",
		--[[disabled = function()
			if (isSpecific) then
				if (not db.validSpell or db.numEvents == 0) then
					return true
				else
					return false
				end
			else
				if (db.numEvents == 0) then
					return true
				else
					return false
				end
			end
		end,]]
		get = function()
			local val
			
			if (db.selected == "empty") then
				val = "Master"
			else
				val = db[db.selected].options.channel
			end
			
			return val
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (isSpecific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								db[k].options.channel = value
							end
						else
							if (v) then
								db[k].options.channel = value
							end
						end
					end
				else
					db[db.selected].options.channel = value
				end
			else
				db[db.selected].options.channel = value
			end
		end,
		values = values,
	}
	
	return channel
end

function Ace3:Volume(db)
	local volume = {
		order = 3,
		type = "range",
		name = "Volume",
		desc = "Adjust the volume",
		min = 0,
		max = 1,
		step = 0.05,
		bigStep = 0.05,
		--[[disabled = function()
			if (isSpecific) then
				if (not db.validSpell or db.numEvents == 0) then
					return true
				else
					return false
				end
			else
				if (db.numEvents == 0) then
					return true
				else
					return false
				end
			end
		end,]]
		get = function()
			local val
			
			if (db.selected == "empty") then
				val = "Master"
			else
				val = db[db.selected].options.channel
			end
			
			return tonumber(format("%.2f",GetCVar("Sound_"..val.."Volume")))
		end,
		set = function(self,value)
			SetCVar("Sound_"..db[db.selected].options.channel.."Volume",value)
		end,
	}
	
	return volume
end

function Ace3:PackSelect(db,tab,isSpecific,noSeparate)
	local packSelect = {
		order = 4,
		type = "select",
		name = "Sound Pack",
		desc = "Choose which sound pack you want to play for crit sounds",
		descStyle = "inline",
		hidden = function()
			if (db.selected ~= "empty") then
				if (db[db.selected].options.source ~= "packs") then
					return true
				else
					return false
				end
			else
				return false
			end
		end,
		disabled = function()
			if (#SCA.library == 0) then
				return true
			else
				if (isSpecific) then
					if (not db.validSpell or db.numEvents == 0) then
						return true
					else
						return false
					end
				elseif (not noSeparate) then
					if (db.numEvents == 0) then
						return true
					else
						return false
					end
				else
					return false
				end
			end
			--[[if (#SCA.library == 0 and (db.numEvents or 1) == 0) then
				return true
			else
				]]
		end,
		get = function()
			if (db.selected ~= "empty") then
				if (db[db.selected].options.source == "packs" and #SCA.library == 0) then
					db[db.selected].packs.value = "empty"
				elseif (#SCA.library > 0 and db[db.selected].packs.path ~= '') then
					for i=1,#SCA.library do
						if (db[db.selected].packs.path == SCA.library[i].path) then
							db[db.selected].packs.value = SCA.library[i].value
						end
					end
				end
				return db[db.selected].packs.value
			else
				return "empty"
			end
		end,
		set = function(self,value)
			for i=1,#SCA.library do
				if (value == SCA.library[i].value) then
					local library = SCA.library[i]

					if (not noSeparate) then
						if (not db.isSeparate) then
							if (value ~= "empty") then
								for k,v in pairs(db.events) do
									if (isSpecific) then
										if (v and Ace3:CheckSpell(k,db.validSpell)) then
											db[k].validAudio = true
											db[k].packs.value = library.value
											db[k].packs.text = library.text
											db[k].packs.path = library.path
											db[k].packs.max = library.max
										end
									else
										if (v) then
											db[k].validAudio = true
											db[k].packs.value = library.value
											db[k].packs.text = library.text
											db[k].packs.path = library.path
											db[k].packs.max = library.max
										end
									end
								end
							else
								for k,v in pairs(db.events) do
									if (isSpecific) then
										if (v and Ace3:CheckSpell(k,db.validSpell)) then
											db[k].validAudio = false
											db[k].packs.value = value
											db[k].packs.text = ''
											db[k].packs.path = ''
											db[k].packs.max = 1
										end
									else
										if (v) then
											db[k].validAudio = false
											db[k].packs.value = value
											db[k].packs.text = ''
											db[k].packs.path = ''
											db[k].packs.max = 1
										end
									end
								end
							end
						else
							if (value ~= "empty") then
								db[db.selected].validAudio = true
								db[db.selected].packs.value = library.value
								db[db.selected].packs.text = library.text
								db[db.selected].packs.path = library.path
								db[db.selected].packs.max = library.max
							else
								db[db.selected].validAudio = false
								db[db.selected].packs.value = value
								db[db.selected].packs.text = ''
								db[db.selected].packs.path = ''
								db[db.selected].packs.max = 1
							end
						end
					else
						if (value ~= "empty") then
							db[db.selected].validAudio = true
							db[db.selected].packs.value = library.value
							db[db.selected].packs.text = library.text
							db[db.selected].packs.path = library.path
							db[db.selected].packs.max = library.max
						else
							db[db.selected].validAudio = false
							db[db.selected].packs.value = value
							db[db.selected].packs.text = ''
							db[db.selected].packs.path = ''
							db[db.selected].packs.max = 1
						end
					end

					break
				end
			end
			
			if (not noSeparate or noSeparate == "partial") then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
		end,
		values = function()
			local packs = {}

			if (#SCA.library > 0) then
				for i=1,#SCA.library do
					packs[SCA.library[i].value] = SCA.library[i].text
				end
				
				packs["empty"] = "Select a pack"
			else
				packs["empty"] = "No packs installed"
			end
			
			return packs
		end,
	}
	
	return packSelect
end

function Ace3:SharedMedia(db,tab,isSpecific,noSeparate)
	local sharedMedia = {
		order = 4,
		type = "select",
		dialogControl = "LSM30_Sound",
		name = "Select Audio",
		desc = "Select an audio file from the SharedMedia library",
		hidden = function()
			if (db.selected == "empty" or db[db.selected].options.source ~= "shared") then
				return true
			else
				return false
			end
		end,
		--[[disabled = function()
			if (isSpecific) then
				if (not db.validSpell or db.numEvents == 0) then
					return true
				else
					return false
				end
			else
				if (db.numEvents == 0) then
					return true
				else
					return false
				end
			end
		end,]]
		get = function()
			if (#db[db.selected].shared.value > 0 and db[db.selected].shared.value ~= "None") then
				db[db.selected].validAudio = true
			else
				db[db.selected].validAudio = false
			end

			return db[db.selected].shared.value
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (specific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								if (value ~= "None") then
									db[k].validAudio = true
								else
									db[k].validAudio = false
								end
								db[k].shared.value = value
							end
						else
							if (v) then
								if (value ~= "None") then
									db[k].validAudio = true
								else
									db[k].validAudio = false
								end
								db[k].shared.value = value
							end
						end
					end
				else
					if (value ~= "None") then
						db[db.selected].validAudio = true
					else
						db[db.selected].validAudio = false
					end
					db[db.selected].shared.value = value
				end
				
				
			else
				if (value ~= "None") then
					db[db.selected].validAudio = true
				else
					db[db.selected].validAudio = false
				end
				db[db.selected].shared.value = value
			end
			
			if (not noSeparate or noSeparate == "partial") then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
		end,
		values = LSM:HashTable("sound"),
	}
	
	return sharedMedia
end

function Ace3:AudioSource(db,values,tab,isSpecific,noSeparate,noTutorial)
	local audioSource = {
		order = 4,
		type = "select",
		name = "Audio Type",
		desc = "Choose whether your custom audio is a file or sound pack",
		hidden = function()
			if (db.selected == "empty" or db[db.selected].options.source ~= "custom") then
				return true
			else
				return false
			end
		end,
		get = function(self)
			if (db[db.selected].custom.customType == "file") then
				if (tab) then
					self.options.args[tab].args.SoundConfig.args.audioLocation.width = "full"
				else
					self.options.args.SoundConfig.args.audioLocation.width = "full"
				end
			else
				if (tab) then
					self.options.args[tab].args.SoundConfig.args.audioLocation.width = "double"
				else
					self.options.args.SoundConfig.args.audioLocation.width = "double"
				end
			end
			
			return db[db.selected].custom.customType
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (isSpecific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								db[k].custom.customType = value
							end
						else
							if (v) then
								db[k].custom.customType = value
							end
						end
					end
				else
					db[db.selected].custom.customType = value
				end
			else
				db[db.selected].custom.customType = value
			end
			
			if (not noTutorial) then
				db.tutorial.flags.customType = true
				Ace3:UpdateTutorial(db)
			end
				
			if (value == "file") then
				if (tab) then
					self.options.args[tab].args.SoundConfig.args.audioLocation.width = "full"
				else
					self.options.args.SoundConfig.args.audioLocation.width = "full"
				end
			else
				if (tab) then
					self.options.args[tab].args.SoundConfig.args.audioLocation.width = "double"
				else
					self.options.args.SoundConfig.args.audioLocation.width = "double"
				end
			end
		end,
		values = values,
	}
	
	return audioSource
end

function Ace3:ToggleSeparate(db,tab,isSpecific,noTutorial)
	local toggleSeparate = {
		order = 6,
		type = "toggle",
		name = "Separate Sound Events",
		disabled = function()
			--[[if (isSpecific) then
				if (not db.validSpell or db.numEvents == 0) then
					return true
				else
					return false
				end
			else]]
				if (db.numEvents > 1) then
					return false
				else
					return true
				end
			--end
		end,
		get = function() 
			return db.isSeparate
		end,
		set = function(self,value)
			db.isSeparate = value

			if (db.isSeparate) then
				for k,_ in pairs(db.events) do
					if (db.selected ~= k) then
						db[k].validAudio = false
						
						db[k].options.source = "packs"
						db[k].options.channel = "Master"
						
						db[k].packs.value = 'empty'
						db[k].packs.text = ''
						db[k].packs.path = ''
						db[k].packs.max = 1
						db[k].packs.isRandom.enable = true
						db[k].packs.isRandom.currentFile = 1
						
						db[k].shared.value = 'None'
						
						db[k].custom.path = "Interface\\AUDIO\\"
						db[k].custom.max = 1
						db[k].custom.customType = 'file'
						db[k].custom.isRandom.enabled = true
						db[k].custom.isRandom.currentFile = 1
					end
				end
			else
				for k,_ in pairs(db.events) do
					if (db.selected ~= k) then
						db[k].validAudio = db[db.selected].validAudio
						
						db[k].options.source = db[db.selected].options.source
						db[k].options.channel = db[db.selected].options.channel

						db[k].packs.value = db[db.selected].packs.value
						db[k].packs.text = db[db.selected].packs.text
						db[k].packs.path = db[db.selected].packs.path
						db[k].packs.max = db[db.selected].packs.max
						db[k].packs.isRandom.enabled = db[db.selected].packs.isRandom.enabled
						db[k].packs.isRandom.currentFile = db[db.selected].packs.isRandom.currentFile
						
						db[k].shared.value = db[db.selected].shared.value
						
						db[k].custom.path = db[db.selected].custom.path
						db[k].custom.max = db[db.selected].custom.max
						db[k].custom.customType = db[db.selected].custom.customType
						db[k].custom.isRandom.enabled = db[db.selected].custom.isRandom.enabled
						db[k].custom.isRandom.currentFile = db[db.selected].custom.isRandom.currentFile
					end
				end
			end

			if (tab) then
				Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific)
			else
				Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific)
			end
			
			if (not noTutorial) then
				db.tutorial.flags.isSeparate = true
				Ace3:UpdateTutorial(db)
			end
		end,
	}
	
	return toggleSeparate
end

function Ace3:IsSCARandom(db,isSpecific,noSeparate)
	local isSCARandom = {
		order = 11,
		type = "toggle",
		name = "Random Order",
		desc = 'Toggle whether files will be played in random or sequential order.',
		disabled = function()
			--[[if (#SCA.library == 0) then
				return true
			else
				if (isSpecific) then
					if (not db.validSpell or db.numEvents == 0) then
						return true
					else
						return false
					end
				elseif (not noSeparate) then
					if (db.numEvents == 0) then
						return true
					else
						return false
					end
				else
					return false
				end
			end]]
			if (#SCA.library == 0) then
				return true
			else
				return false
			end
		end,
		hidden = function()
			if (db.selected == "empty" or db[db.selected].options.source ~= "packs") then
				return true
			else
				return false
			end
		end,
		get = function()
			return db[db.selected].packs.isRandom.enabled
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (isSpecific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								db[k].packs.isRandom.enabled = value
							end
						else
							if (v) then
								db[k].packs.isRandom.enabled = value
							end
						end
					end
				else
					db[db.selected].packs.isRandom.enabled = value
				end
			else
				db[db.selected].packs.isRandom.enabled = value
			end
		end,
		width = "double",
	}
	
	return isSCARandom
end

function Ace3:AudioLocation(db,isSpecific,noSeparate,noTutorial)
	local audioLocation = {
		type = "input",
		order = 8,				
		name = "Audio Path",
		desc = "Audio file/path location.\n\nFile extensions not necessary.\n\nOnly |cFF00FF00MP3|r and |cFF00FF00OGG|r files are supported!",
		
		hidden = function()
			if (db.selected == "empty" or db[db.selected].options.source ~= "custom") then
				return true
			else
				return false
			end
		end,
		get = function()
			return Ace3:VerifyCustomPath(db[db.selected].custom.path,db[db.selected])
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (isSpecific) then
							if (v and Ace3:CheckSpell(k,db.validSpell)) then
								db[k].custom.path = Ace3:VerifyCustomPath(value,db[k])
							end
						else
							if (v) then
								db[k].custom.path = Ace3:VerifyCustomPath(value,db[k])
							end
						end
					end
				else
					db[db.selected].custom.path = Ace3:VerifyCustomPath(value,db[db.selected])
				end
			else
				db[db.selected].custom.path = Ace3:VerifyCustomPath(value,db[db.selected])
			end
			
			if (not noTutorial) then
				db.tutorial.flags.audioLocationFile = true
				db.tutorial.flags.audioLocationPack = true
				Ace3:UpdateTutorial(db)
			end
		end,
	}
	
	return audioLocation
end

function Ace3:NumFiles(db,errorName,tab,isSpecific,noSeparate,noTutorial)
	local numFiles = {
		type = "input",
		order = 9,
		name = "Number of files",
		desc = "Indicate the amount of files in the sound pack",
		hidden = function()
			if (db.selected == "empty" or db[db.selected].custom.customType == "file" or db[db.selected].options.source ~= "custom") then
				return true
			else
				return false
			end
		end,
		get = function()
			return tostring(db[db.selected].custom.max)
		end,
		set = function(self,value)
			local errors
			
			if (tab) then
				errors = self.options.args[tab].args.errors.args.messages
			else
				errors = self.options.args.errors.args.messages
			end
			
			if (type(tonumber(value)) == "number") then
				if (not noSeparate) then
					if (not db.isSeparate) then
						for k,v in pairs(db.events) do
							if (isSpecific) then
								if (v and Ace3:CheckSpell(k,db.validSpell)) then
									db[k].custom.max = tonumber(value)
								end
							else
								if (v) then
									db[k].custom.max = tonumber(value)
								end
							end
						end
					else
						db[db.selected].custom.max = tonumber(value)
					end
				else
					db[db.selected].custom.max = tonumber(value)
				end
				errors.name = " "
				SCA.temp.errors[errorName] = nil
			else
				errors.name = "Number of Files: |cFFFF5F99Numeric values only.|r"
				
				SCA.temp.errors[errorName] = true
			end
			
			if (not noTutorial) then
				db.tutorial.flags.numFiles = true
				Ace3:UpdateTutorial(db)
			end
		end,
	}
	
	return numFiles
end

function Ace3:TestSound(db,errorName,tab,isSpecific,noSeparate,noTutorial)
	local test = {
		order = 10,
		type = "execute",
		name = "Test",
		hidden = function()
			if (db.selected == "empty" or db[db.selected].options.source ~= "custom") then
				return true
			else
				return false
			end
		end,
		func = function(self)
			local errors
			
			if (tab) then
				errors = self.options.args[tab].args.errors.args.messages
			else
				errors = self.options.args.errors.args.messages
			end
			
			if (db[db.selected].custom.customType == "pack") then
				--print("PATH: "..db[db.selected].custom.path..tostring(Ace3:GetSoundFileNumber(db[db.selected],"custom"))..".mp3")
				if (not PlaySoundFile(db[db.selected].custom.path..tostring(Ace3:GetSoundFileNumber(db[db.selected],"custom"))..".mp3",db[db.selected].options.channel) and not PlaySoundFile(db[db.selected].custom.path..tostring(Ace3:GetSoundFileNumber(db[db.selected],"custom"))..".ogg",db[db.selected].options.channel)) then
					errors.name = "Custom Audio: |cFFFF5F99Audio file not found!.|r"
					SCA.temp.errors[errorName] = true
					
					if (not db[db.selected].custom.isRandom.enabled) then
						db[db.selected].custom.isRandom.currentFile = db[db.selected].custom.isRandom.currentFile - 2
					end

					if (not noSeparate) then
						if (not db.isSeparate) then
							for k,v in pairs(db.events) do
								if (isSpecific) then
									if (v and Ace3:CheckSpell(k,db.validSpell)) then
										db[k].validAudio = false
									end
								else
									if (v) then
										db[k].validAudio = false
									end
								end
							end
						else
							db[db.selected].validAudio = false
						end
					else
						db[db.selected].validAudio = false
					end
				else
					errors.name = " "
					SCA.temp.errors[errorName] = nil
					
					if (not db[db.selected].custom.isRandom.enabled) then
						db[db.selected].custom.isRandom.currentFile = db[db.selected].custom.isRandom.currentFile - 1
					end

					if (not noSeparate) then
						if (not db.isSeparate) then
							for k,v in pairs(db.events) do
								if (isSpecific) then
									if (v and Ace3:CheckSpell(k,db.validSpell)) then
										db[k].validAudio = true
									end
								else
									if (v) then
										db[k].validAudio = true
									end
								end
							end
						else
							db[db.selected].validAudio = true
						end
					else
						db[db.selected].validAudio = true
					end
				end
			else
				if (not PlaySoundFile(db[db.selected].custom.path..".mp3",db[db.selected].options.channel) and not PlaySoundFile(db[db.selected].custom.path..".ogg",db[db.selected].options.channel)) then
					errors.name = "Custom Audio: |cFFFF5F99Audio file not found!|r"
					SCA.temp.errors[errorName] = true
					
					if (not noSeparate) then
						if (not db.isSeparate) then
							for k,v in pairs(db.events) do
								if (isSpecific) then
									if (v and Ace3:CheckSpell(k,db.validSpell)) then
										db[k].validAudio = false
									end
								else
									if (v) then
										db[k].validAudio = false
									end
								end
							end
						else
							db[db.selected].validAudio = false
						end
					else
						db[db.selected].validAudio = false
					end
				else
					errors.name = " "
					SCA.temp.errors[errorName] = nil
					
					if (not noSeparate) then
						if (not db.isSeparate) then
							for k,v in pairs(db.events) do
								if (isSpecific) then
									if (v and Ace3:CheckSpell(k,db.validSpell)) then
										db[k].validAudio = true
									end
								else
									if (v) then
										db[k].validAudio = true
									end
								end
							end
						else
							db[db.selected].validAudio = true
						end
					else
						db[db.selected].validAudio = true
					end
				end
			end
			
			if (not noTutorial) then
				Ace3:UpdateTutorial(db)
			end
			
			if (not noSeparate or noSeparate == "partial") then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
		end,
	}
	
	return test
end

function Ace3:IsCustomRandom(db,noSeparate)
	local isCustomRandom = {
		order = 11,
		type = "toggle",
		name = "Random Order",
		desc = 'Toggle whether files will be played in random or sequential order.',
		hidden = function()
			if (db.selected == "empty" or db[db.selected].custom.customType == "file" or db[db.selected].options.source ~= "custom") then
				return true
			else
				return false
			end
		end,
		get = function()
			return db[db.selected].custom.isRandom.enabled
		end,
		set = function(self,value)
			if (not noSeparate) then
				if (not db.isSeparate) then
					for k,v in pairs(db.events) do
						if (v) then
							db[k].custom.isRandom.enabled = value
						end
					end
				else
					db[db.selected].custom.isRandom.enabled = value
				end
			else
				db[db.selected].custom.isRandom.enabled = value
			end
		end,
		width = "double",
	}
	
	return isCustomRandom
end

function Ace3:SoftReset(db,tab,isSpecific,noSeparate)
	local reset = {
		order = 12,
		type = "execute",
		name = "Reset",
		disabled = function()
			if (db.selected ~= "empty" and db[db.selected].validAudio) then
				return false
			else
				return true
			end
		end,
		desc = function()
			return "Resets the sound for: |cFF00FF00"..db[db.selected].label.."|r"
		end,
		func = function(self)
			Ace3:ResetDB(db)
			
			if (not noSeparate or noSeparate == "partial") then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
		end,
		width = "half",
	}
	
	return reset
end

function Ace3:HardReset(db,tab,isSpecific,noSeparate)
	local reset = {
		order = 11,
		type = "execute",
		name = function()
			if (IsShiftKeyDown()) then
				return "Reset Everything"
			else
				return "Reset All Sounds"
			end
		end,
		hidden = function()
			if (isSpecific) then
				if (SCA.temp.addSound) then
					return false
				else
					return true
				end
			else
				return false
			end
		end,
		desc = function()
			if (IsShiftKeyDown()) then
				return "Resets all settings and sound configurations"
			else
				return "Resets all sound configurations.\n\n|cFF00FF00Holding SHIFT while clicking will reset EVERYTHING in this tab."
			end
		end,
		disabled = function()
			if (IsShiftKeyDown()) then
				return false
			else
				for k,v in pairs(db.events) do
					if (db[k].validAudio) then
						return false
					end
				end
				
				return true
			end
		end,
		func = function(self)
			Ace3:ResetDB(db,IsShiftKeyDown())
			
			if (noSeparate) then
				if (tab) then
					Ace3:UpdateSeparateSoundMessage(db,self.options.args[tab],isSpecific,noSeparate)
				else
					Ace3:UpdateSeparateSoundMessage(db,self.options,isSpecific,noSeparate)
				end
			end
		end,
	}
end

--[[-------------------------------------------
	
	DAMAGE & HEALING SETUP TOOLS
	
---------------------------------------------]]
function Ace3:Threshold(db,event,title,errorName,tab)
	local threshold = {
		order = 4,
		type = "input",
		name = title.." Threshold",
		desc = "Restrict the audio queues to a specific "..lower(title).." threshold\n\n|cFFFF5F99Digit group is automatically applied; do not enter commas.|r",
		get = function() 
			return BreakUpLargeNumbers(tonumber(db[event].options.threshold))
		end,
		set = function(self,value)
			local errors
			
			if (tab) then
				errors = self.options.args[tab].args.errors.args.messages
			else
				errors = self.options.args.errors.args.messages
			end
			
			if (type(tonumber(value)) == "number") then
				db[event].options.threshold = tonumber(value)
				errors.name = " "
				SCA.temp.errors[errorName] = nil
			else
				local isComma = find(value,",")
				
				if (isComma) then
					errors.name = title.." Threshold: |cFFFF5F99Do not enter commas; they are added automatically.|r"
				else
					errors.name = title.." Threshold: |cFFFF5F99Numeric values only.|r"
				end
				
				SCA.temp.errors[errorName] = true
			end
		end,
	}
	
	return threshold
end

function Ace3:CritSelect(db,event,triggerList)
	local critSelect = {
		order = 5,
		type = "select",
		name = "Crit / Non-Crit",
		desc = "Choose which sound pack you want to play for crit sounds",
		descStyle = "inline",
		get = function()
			return db[event].options.trigger
		end,
		set = function(self,value)
			db[event].options.trigger = value
		end,
		values = triggerList,
	}
	
	return critSelect
end